/**
 * Library to control Sky box via USB 'sky controller'
 *
 * Copyright (C) 2005 Joseph Heenan. All rights reserved.
 * Copyright (C) 2005 Neil Goodman. All rights reserved.
 *
 * This software is provided "as-is", without any express or implied warranty.
 * In no event shall the author be held liable for any damages arising from
 * the use of this software. 
 * 
 * $Id: sky-control-lib.c,v 1.3 2009/07/06 15:42:16 joseph Exp $
 * 
 */

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <windows.h>

#include "setupapi.h"
#include "hidsdi.h"

#include "sky-control-lib.h"

/* usb vendor and product id for sky controller */
#define VENDOR_ID  0x6666 
#define PRODUCT_ID 0xf100

struct sky_usb
{
    HANDLE             handle;
    HANDLE             device;
    int                debug;
    int                timeout;
};

static GUID                         gs_HidGuid;
static int                          gs_DeviceIndex;
static HANDLE                       gs_DeviceInfo;
static SP_DEVICE_INTERFACE_DATA     gs_InterfaceData;

int sky_usb_initialise(void)
{
    return 0;
}

void sky_usb_finalise(void)
{
}

sky_usb *sky_usb_create(int debug, int timeout)
{
    sky_usb *sky = malloc(sizeof(*sky));
    if (sky == NULL)
        return NULL;
    
    sky->handle            = INVALID_HANDLE_VALUE;
    sky->debug             = debug;
    sky->timeout           = timeout;

    return sky;
}

void sky_usb_destroy(sky_usb *sky)
{
    if (sky == NULL)
        return;

    if (sky->handle != INVALID_HANDLE_VALUE)
        CloseHandle(sky->handle);
    
    free(sky);
}

static void OpenFindHID() 
{
    HidD_GetHidGuid(&gs_HidGuid);

    gs_DeviceInfo  = SetupDiGetClassDevs(&gs_HidGuid, NULL, NULL, DIGCF_PRESENT | DIGCF_INTERFACEDEVICE);
    gs_DeviceIndex = 0;
}

static SP_DEVICE_INTERFACE_DATA *FindNextHID() 
{
    gs_InterfaceData.cbSize = sizeof(gs_InterfaceData);

    if (SetupDiEnumDeviceInterfaces(gs_DeviceInfo, 0, &gs_HidGuid, gs_DeviceIndex++, &gs_InterfaceData) == 0)
    {
        //  No more devices
        return (NULL);
    }

    //  Return the device interface data
    return (&gs_InterfaceData);
}

static void CloseFindHID() 
{
    SetupDiDestroyDeviceInfoList(gs_DeviceInfo);
    gs_DeviceInfo = NULL;
}

//  .................................................................HID Detail

static PSP_DEVICE_INTERFACE_DETAIL_DATA GetHIDDetail(SP_DEVICE_INTERFACE_DATA* pInterfaceData) 
{
    PSP_DEVICE_INTERFACE_DETAIL_DATA pDetailData;
    ULONG Length = 0;

    //  Determine the size of the detail data
    SetupDiGetDeviceInterfaceDetail(gs_DeviceInfo, pInterfaceData, NULL, 0, &Length, NULL);

    //  Allocate a buffer for the detail data
    pDetailData = (PSP_DEVICE_INTERFACE_DETAIL_DATA) malloc(Length);
    pDetailData->cbSize = sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA);

    //  Obtain the device interface's detail data
    SetupDiGetDeviceInterfaceDetail(gs_DeviceInfo, pInterfaceData, pDetailData, Length, NULL, NULL);

    return (pDetailData);
}

static void ReleaseHIDDetail(PSP_DEVICE_INTERFACE_DETAIL_DATA pDetailData) 
{
    free(pDetailData);
}

//  ...........................................................................

/**
 * Attempt to find the sky controller usb device
 * 
 * Scans through all the devices on all the usb busses and finds the
 * 'device_num'th controller.
 * 
 * @param device_num 0 for first device, 1 for second, etc
 */
int sky_usb_finddevice(sky_usb *sky, int device_num)
{
    PSP_DEVICE_INTERFACE_DETAIL_DATA pDetailData;
    int device_count = 0;

    assert(sky != NULL);
    assert(sky->handle == INVALID_HANDLE_VALUE);
    assert(device_num >= 0 && device_num < 255);

    //  Iterate over every present device
    OpenFindHID();

    while (sky->handle == INVALID_HANDLE_VALUE)
    {
        SP_DEVICE_INTERFACE_DATA* pInterfaceData = FindNextHID();

        if (pInterfaceData == NULL)
        {
            break;
        }

        //  Obtain the device's details
        pDetailData = GetHIDDetail(pInterfaceData);

        if (pDetailData != NULL)
        {
            //  Open a handle to the device
            HANDLE hDevice = CreateFile(pDetailData->DevicePath, GENERIC_READ | GENERIC_WRITE,
                FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);

            if (hDevice != NULL)
            {
                //  Read the device attributes
                HIDD_ATTRIBUTES Attributes;
                Attributes.Size = sizeof(Attributes);
                HidD_GetAttributes(hDevice, &Attributes);

                if (sky->debug)
                {
                    fprintf(stderr, "Found device: %04x-%04x\n",
                           Attributes.VendorID,
                           Attributes.ProductID);
                }

                //  Is this the Sky controller device?
                if ((Attributes.VendorID  == VENDOR_ID) &&
                    (Attributes.ProductID == PRODUCT_ID))
                {
                    if (sky->debug)
                        fprintf(stderr, "Found sky control num %d\n", device_count);
                    //  Sky controller detected
                    if (device_count != device_num)
                    {
                        device_count++;
                    }
                    else
                    {
                        //  Get the device's capabilities
                        HIDP_CAPS Capabilities;
                        PHIDP_PREPARSED_DATA PreparsedData;
                        HidD_GetPreparsedData(hDevice, &PreparsedData);
                        HidP_GetCaps(PreparsedData, &Capabilities);
                        HidD_FreePreparsedData(PreparsedData);

                        //  Open the usable handle to the device
                        sky->handle = CreateFile(pDetailData->DevicePath, GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);
                        if (sky->handle == INVALID_HANDLE_VALUE)
                        {
                            fprintf(stderr, "Failed to open device (%d)\n", GetLastError());
                        }
                        sky->device = hDevice;
                        hDevice = INVALID_HANDLE_VALUE;
                    }
                }

                //  Release the device handle
                if (hDevice != INVALID_HANDLE_VALUE)
                    CloseHandle(hDevice);
            }

            //  Release the detail data
            ReleaseHIDDetail(pDetailData);
        }
    }

    //  Release the HID enumerator
    CloseFindHID();

    if (sky->handle == INVALID_HANDLE_VALUE)
    {
        fprintf(stderr, "sky_usb_finddevice(%d) failed to find device %04x-%04x\n",
                device_count, VENDOR_ID, PRODUCT_ID);
        return -1;
    }

    return 0;
}

void sky_usb_printinfo(sky_usb *sky)
{
    /* not sure how to do this on windows - this doesn't seem to work */
    wchar_t string[512];
    int  ret;
    
    assert(sky != NULL);
    assert(sky->handle != INVALID_HANDLE_VALUE);
    
    ret = HidD_GetManufacturerString(sky->device,
                                     string, sizeof(string));
    if (ret != 0)
    {
        char str[512];
        char *ptr;
        wchar_t *x;
        for (ptr=str,x=string; *string != 0;)
        {
            *ptr++  = (char)*x++;
        }
        *ptr = 0;
        fprintf(stderr, "- Manufacturer : %s\n", ptr);
    }
    else
        fprintf(stderr, "- Unable to fetch manufacturer string\n");
/*    
    if (sky->device->descriptor.iProduct) 
    {
        ret = usb_get_string_simple(sky->handle,
                                    sky->device->descriptor.iProduct,
                                    tmp_string, sizeof(tmp_string));
        if (ret > 0)
          fprintf(stderr, "- Product : %s\n", tmp_string);
        else
          fprintf(stderr, "- Unable to fetch product string\n");
    }
    
                    
    if (sky->device->descriptor.iSerialNumber) 
    {
        ret = usb_get_string_simple(sky->handle,
                                    sky->device->descriptor.iSerialNumber,
                                    tmp_string, sizeof(tmp_string));
        if (ret > 0)
          fprintf(stderr, "- Serial Number: %s\n", tmp_string);
        else
          fprintf(stderr, "- Unable to fetch serial number string\n");
    }
*/
}

int sky_usb_sendcommand(sky_usb *sky, int output, int command)
{
    BOOL success;
    unsigned char buf[9];
    unsigned long bytes = 0;
    
    assert(sky != NULL);
    assert(sky->handle != INVALID_HANDLE_VALUE);
    assert(output == 0 || output == 1);
    assert(command >= 0 && command < 0xffff);
    
    memset(buf, 122, sizeof(buf));
    buf[0] = 0;
    buf[1] = output == 0 ? '0' : '1';
    buf[2] = command >> 8;
    buf[3] = command & 0xff;
    
    if (sky->debug)
    {
        fprintf(stderr, "sending cmd %04x to output %d: bytes %d, %d, %d\n",
               command, output+1, buf[1], buf[2], buf[3]);
    }
        
    success = WriteFile(sky->handle, buf, sizeof(buf), &bytes, NULL);
    if (!success || bytes != sizeof(buf))
    {
        fprintf(stderr, "WriteFile failed (%d / %d) : %d\n", success, bytes,GetLastError());
        return -1;
    }
    
    return 0;
}

int sky_usb_sendconfig(sky_usb *sky, int output, sky_config *config)
{
    unsigned char buf[9];
    int ret;
    unsigned long bytes = 0;
    
    assert(sky != NULL);
    assert(sky->handle != INVALID_HANDLE_VALUE);
    assert(output == 0 || output == 1);
    assert(config != NULL);
    
    memset(buf, 122, sizeof(buf));
    buf[0] = 0;
    buf[1] = output == 0 ? 's' : 't';
    buf[2] = config->repeatdelay;
    buf[3] = config->interkeydelay;
    buf[4] = config->samekeydelay;
    buf[5] = config->interkeydelayhigh;
    buf[6] = config->interkeydelaylow;
    buf[7] = config->interbitdelayhigh;
    buf[8] = config->interbitdelaylow;
    
    ret = WriteFile(sky->handle, buf, sizeof(buf), &bytes, NULL);
    if (ret != 0 || bytes != sizeof(buf))
    {
        fprintf(stderr, "WriteFile failed (%d)\n", GetLastError());
        return -1;
    }

    memset(buf, 122, sizeof(buf));
    buf[0] = 0;
    buf[1] = output == 0 ? 'u' : 'v';
    buf[2] = config->repeatcount;

    ret = WriteFile(sky->handle, buf, sizeof(buf), &bytes, NULL);
    if (ret != 0 || bytes != sizeof(buf))
    {
        fprintf(stderr, "WriteFile failed (%d)\n", GetLastError());
        return -1;
    }
    
    return 0;
}
